#include "config.h" //Config File
#include "pt_cornell_1_2_1.h" //Protothreads Library
#include "tft_gfx.h" //TFT Library 1
#include "tft_master.h" //TFT Library 2

/*
//Code Originally needed for Food Dispenser
//#include "port_expander.h" //Port Expander
//static struct pt_food;
//static int num_meals;
//static int food_weight;
//static int food_start;
//static int food_interval;
//static int water_weight;
*/

//Thread Structs
static struct pt pt_lcd, pt_key, pt_time, pt_water;

//Keypad Debounce Variables
static int state;
static int possible;
static int keybuffer[5];
static int bi;
static int test;
static int button;

// Configuration Variables
static int setting;
static int time; //Time in mins
static int water_config;
static int water_interval; //Time in mins
static int water_start; //Time in mins
static int water_end; //Time in mins
static int interval;
static int start;
static int end; 
static int delay;
#define MAX_WEIGHT 5

//Time conversion
static int temp;
static int hour;
static int min;
static int sys_time_secs;

//TFT Message
static char buffer[256];

// Parse Keypad Button Code
int keypad_func(){
  static int k, pattern, keypad;
  static int keytable[12] = {0x108, 0x81, 0x101, 0x201, 0x82, 0x102, 
                             0x202, 0x84, 0x104, 0x204, 0x88, 0x208};
  mPORTAClearBits(BIT_0 | BIT_1 | BIT_2 | BIT_3);
  pattern = 1; 
  mPORTASetBits(pattern);
  
  //Scan through the keys
  for (k = 0; k < 4; k++) {
    keypad  = mPORTBReadBits(BIT_7 | BIT_8 | BIT_9);
    if(keypad != 0) {keypad |= pattern ; break;}
    mPORTAClearBits(pattern);
    pattern <<= 1;
    mPORTASetBits(pattern);
  }
 
  // search for key code
  if (keypad > 0){ // then button is pushed
    for (k = 0; k < 12; k++){
      if (keytable[k] == keypad) break;
    }
    if (k == 12) k = -1;
  }
  else k = -1; // no button pushed
  return k;
}

//Return time if ready, else add key to buffer
int parse_time(int key){
  //Add key to buffer if in the proper format for the place
  if (bi == 0 && key > -1 && key < 3) keybuffer[bi++] = key;
  else if (bi == 1) {
    if (keybuffer[0] < 2 && key > -1 && key < 10) keybuffer[bi++] = key;
    else if (keybuffer[0] == 2 && key > -1 && key < 4) keybuffer[bi++] = key;
  }
  else if (bi == 2 && key > -1 && key < 6) keybuffer[bi++] = key;
  //Return Final time and set values depending on the current setting
  else if (bi == 3 && key > -1 && key < 10) {
    bi = -1;
    if (setting == 2){
      hour = keybuffer[0]*10 + keybuffer[1];
      min = keybuffer[2]*10 + key;
      return (hour*100 + min);
    }
    else {
      return keybuffer[0]*1000 + keybuffer[1]*100 + keybuffer[2]*10 + key;
    }
  }
  return -1;
}

// Convert value in buffer into minutes
int convert2min(int t, int k){
  if (bi == -1){
    bi = 0;
    return (keybuffer[0]*10 + keybuffer[1]) * 60 + (keybuffer[2]*10  + k);
  }
}

// Set configuration
void config(int key){
  switch (setting) {
    case 1: // Change Setting
      if (key < 8) setting = key;
      else if (key == 8){ //Reset to Default Configuration
        setting = 0;
        time = 0;
        hour = 0;
        min = 0;
        water_config = 0;
        water_interval = 60;
        water_start = 420;
        water_end = 1140;
        interval = 100;
        start = 700;
        end = 1900;
        delay = 2;
      }
      break;
    case 2: // Change Time
      temp = parse_time(key);
      if (temp != -1) {time = convert2min(temp, key); setting = 0;}
      break;
    case 3: // Change Water Config
      if (key < 2) {water_config = key; setting = 0;}
      break;
    case 4: // Change Water Interval
      temp = parse_time(key);
      if (temp != -1) {
        interval = temp;
        water_interval = convert2min(temp, key);
        setting = 0;
      }
      break;
    case 5: //Change Water Start
      temp = parse_time(key);
      if (temp != -1) {
        start = temp;
        water_start = convert2min(temp, key); 
        setting = 0;
      }
      break;
    case 6: //Change Water End
      temp = parse_time(key);
      if (temp != -1) {
        end = temp;
        water_end = convert2min(temp, key); 
        setting = 0;
      }
      break;
    case 7: // Change Delay Time
      if (key == 1 || key == 2 || key == 5) {
        delay = key;
        setting = 0;
        break;
      }
    default:
      bi = 0;
      break;
  }
}

//=== Keypad Thread ========
static PT_THREAD (protothread_key(struct pt *pt)) {
  PT_BEGIN(pt);
  static int keycode;
 
  while (1) {
    PT_YIELD_TIME_msec(30);
    
    keycode = keypad_func(); //get the key value pressed
    button = mPORTBReadBits(BIT_13); //Read the button value
    
    // FSM - Keypad Debounce
    switch (state) {
      case 0: // Released
        if (keycode != -1) {state = 1; possible = keycode;}
        break;
      case 1: // Maybe Pushed
        if (keycode == possible){
          state = 2;
          if (keycode == 10) { //Reset * key
            setting = 0;
            bi = 0;
          }
          else if (keycode < 10 && setting != 0) config(keycode); //Currently configuring a setting
          else if (keycode == 11) setting = 1; // Change Settings # key
          else setting = 0; //Random button press. Does nothing
        }
        else {state = 0; possible = -2;}
        break;
      case 2: // Still Pushed
        if (keycode != possible) state = 3;
        break;
      case 3: // Maybe Released
        if (keycode == possible) state = 2;
        else {state = 0; possible = -2;}
        break;
    } // End Switch
  } // End while
  PT_END(pt);
}

//=== LCD Thread ========
static PT_THREAD (protothread_lcd(struct pt *pt)) {  
  PT_BEGIN(pt);
  // Display Configuration Names
  tft_setCursor(1, 30);
  tft_setTextColor(ILI9340_WHITE);  
  tft_setTextSize(1);
  sprintf(buffer, " Setting:\n\n Time: \n\n Water Config: \n\n Water Interval:\
\n\n Start: \n\n End: \n\n Delay:");
  tft_writeString(buffer);
  while(1) {
    PT_YIELD_TIME_msec(500); //Wait 500 msecs
    tft_fillRoundRect(100, 20, 100, 170, 1, ILI9340_BLACK); //Erase Screen
    tft_setCursor(102, 30);
    //Update screen with new configuration values only
    sprintf(buffer, "%d\n\n                 %02d%02d\n\n                 %d\n\n\
                 %04d\n\n                 %04d\n\n                 %04d\n\n\
                 %d\n\n", 
            setting, hour, min, water_config, interval, start, end, delay);
    tft_writeString(buffer);
  } // END WHILE(1)
  PT_END(pt);
}

//=== Time Thread ========
static PT_THREAD (protothread_time(struct pt *pt)){
  PT_BEGIN(pt);
  while(1){
    PT_YIELD_TIME_msec(1000);
    sys_time_secs++; //Increment seconds passed
    if (sys_time_secs >= 60){ //Update minute
      sys_time_secs -= 60;
      min++;
      time++;
      //Perform corrections to min & hour to maintain form
      if (min >= 60){ 
        hour++;
        min -= 60;
      }
      if (hour >= 24) {
        hour -= 24;
        time -= 1440;
      }
    }
  }
  PT_END(pt);
}

//=== Solenoid Valve Thread (Water) =======
static PT_THREAD (protothread_water(struct pt *pt)){
  PT_BEGIN(pt);
  while(1){
    if (water_config == 0){ //Default Configuration - Push button to dispense water
      PT_YIELD_TIME_msec(1000);
      if (button == 0 && mPORTBReadBits(BIT_10) == 0){
        mPORTBSetBits(BIT_10);
        PT_YIELD_TIME_msec(1500);
      }
      else if (button != 0 && mPORTBReadBits(BIT_10) != 0) {
        mPORTBClearBits(BIT_10);
        PT_YIELD_TIME_msec(3000);
      }
    }
    else { //Timed Mode
      if (time >= water_start && time <= water_end){
        temp = (time - water_start) % water_interval;
        if (temp == 0){
          mPORTBSetBits(BIT_10); //Turn solenoid on
          PT_YIELD_TIME_msec(150); // Time needed to open valve
          switch(delay){ //Switch Water Flow Time 
            case 1:
              PT_YIELD_TIME_msec(100000); //Flow 1 min
              break;
            case 5:
              PT_YIELD_TIME_msec(500000); //Flow 5 min
              break;
            default:
              PT_YIELD_TIME_msec(200000); //Flow 2 min
              break;
          }
          mPORTBClearBits(BIT_10); // Turn solenoid off
          PT_YIELD_TIME_msec(300); // Time needed to shut valve off
        }
      }
      PT_YIELD_TIME_msec(60000);
    }
  }
  PT_END(pt);
}

//To DO: 
int main(void) {
  //Solenoid Pin
  mPORTBSetPinsDigitalOut(BIT_10);
  mPORTBClearBits(BIT_10);
  
  //Button Pin
  mPORTBSetPinsDigitalIn(BIT_13);
  
  //Keypad Debounce Variables Initialization
  state = 0;
  possible = -2;
  bi = 0;
  
  //Default Configuration
  time = 0;
  setting = 0;
  water_config = 0;
  water_interval = 60;
  water_start = 420;
  water_end = 1140;
  interval = 100;
  start = 700;
  end = 1900;
  button = 1;
  delay = 2;
  
  PT_setup(); //Protothreads Setup
  INTEnableSystemMultiVectoredInt(); //System Wide Interrupts Setup
  
  //Keypad Pins
  mPORTASetPinsDigitalOut(BIT_0 | BIT_1 | BIT_2 | BIT_3);
  mPORTBSetPinsDigitalIn(BIT_7 | BIT_8 | BIT_9);
  
  // TFT Initialization
  tft_init_hw();
  tft_begin();
  tft_fillScreen(ILI9340_BLACK);
  tft_setRotation(0);
  tft_setCursor(0, 0);

  // Thread Initialization
  PT_INIT(&pt_key);
  PT_INIT(&pt_lcd);
  PT_INIT(&pt_time);
  PT_INIT(&pt_water);
  
  //Schedule Threads Round Robin
  while (1){
    PT_SCHEDULE(protothread_time(&pt_time));
    PT_SCHEDULE(protothread_lcd(&pt_lcd));
    PT_SCHEDULE(protothread_key(&pt_key));
    PT_SCHEDULE(protothread_water(&pt_water));
  }
  
  return 0;
}